from typing import Dict

from report.symbolhistory.symbolhistorycontextholding import SymbolHistoryContextHolding


class SymbolHistoryContextSnapshot(object):

    def __init__(
        self,
        profit: float,
        stock_holding: SymbolHistoryContextHolding,
        option_sell_holding: Dict[str, SymbolHistoryContextHolding],
        option_buy_holding: Dict[str, SymbolHistoryContextHolding],
    ) -> None:
        self._profit = profit
        self._stock_holding = stock_holding
        self._option_sell_holding = option_sell_holding
        self._option_buy_holding = option_buy_holding

    @property
    def profit(self) -> float:
        return self._profit

    @property
    def stock_holding(self) -> SymbolHistoryContextHolding:
        return self._stock_holding

    @property
    def option_buy_holding(self) -> Dict[str, SymbolHistoryContextHolding]:
        return self._option_buy_holding

    @property
    def option_sell_holding(self) -> Dict[str, SymbolHistoryContextHolding]:
        return self._option_sell_holding



from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from robinhood.lib.account.optionevent.optioneventequitycomponent import option_event_equity_component_from_data
from robinhood.lib.account.optionevent.optioneventequitycomponent import OptionEventEquityComponent
from robinhood.lib.helpers import parse_date
from robinhood.lib.helpers import parse_datetime_with_microsecond
from robinhood.lib.helpers import parse_optional_float
from robinhood.lib.marketplace.optioninstrument import option_instrument_from_url
from robinhood.lib.marketplace.optioninstrument import OptionInstrument


class OptionEvent(object):

    def __init__(
        self,
        event_id: str,
        event_type: str,
        event_date: datetime,
        state: str,
        account_url: str,
        position_url: str,
        option_instrument_url: str,
        chain_id: str,
        equity_components: List[OptionEventEquityComponent],
        direction: str,
        underlying_price: Optional[float],
        total_cash_amount: float,
        quantity: float,
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._event_id = event_id
        self._event_type = event_type
        self._event_date = event_date
        self._state = state
        self._account_url = account_url
        self._position_url = position_url
        self._option_instrument_url = option_instrument_url
        self._chain_id = chain_id
        self._equity_components = equity_components
        self._direction = direction
        self._underlying_price = underlying_price
        self._total_cash_amount = total_cash_amount
        self._quantity = quantity
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def state(self) -> str:
        return self._state

    @property
    def event_type(self) -> str:
        return self._event_type

    @property
    def option_instrument(self) -> OptionInstrument:
        return option_instrument_from_url(url=self._option_instrument_url)

    @property
    def quantity(self) -> float:
        return self._quantity

    @property
    def direction(self) -> str:
        return self._direction

    @property
    def created_at(self) -> datetime:
        return self._created_at


def option_event_from_data(data: Dict[str, Any]) -> OptionEvent:
    equity_components = [
        option_event_equity_component_from_data(data=row)
        for row in data['equity_components']
    ]
    option_event = OptionEvent(
        event_id=data['id'],
        event_type=data['type'],
        event_date=parse_date(data['event_date']),
        state=data['state'],
        account_url=data['account'],
        position_url=data['position'],
        option_instrument_url=data['option'],
        chain_id=data['chain_id'],
        equity_components=equity_components,
        direction=data['direction'],
        underlying_price=parse_optional_float(data['underlying_price']),
        total_cash_amount=float(data['total_cash_amount']),
        quantity=float(data['quantity']),
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )
    return option_event



from report.optionsellscreener.optionsellscreener import OptionSellScreener


report = OptionSellScreener(
    exp_date='2021-01-15',
    option_type='put',
    min_chance_of_profit=0.85,
    min_fund_efficiency=0.0075,
)


def test_get_all_equity_instrument_ids() -> None:
    ids = report._get_all_equity_instrument_ids()
    assert len(ids) > 0
    for i in ids:
        assert len(i) == 36


def test_format_data_row() -> None:
    formatted = report._format_data_row([
        'Some Symbol',
        1.23,
        2.34,
        0.45523,
        0.892323,
        2.45,
        3.45,
        4.56,
        123.89,
        0.00875,
        13456,
    ])
    assert formatted == [
        'Some Symbol',
        '$1.23',
        '$2.34',
        '45.52%',
        '89.23%',
        '$2.45',
        '$3.45',
        '$4.56',
        '$123.89',
        '0.88%',
        '13,456',
    ]


def test_gen_ascii_tbl_data() -> None:
    data = report._gen_ascii_tbl_data()
    assert len(data) > 0
    for row in data:
        assert len(row) == 11



from collections import defaultdict
from collections import OrderedDict
from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from asciitree import LeftAligned
from termcolor import colored
from terminaltables import AsciiTable

from data.watchlist import ALL_REVERSED
from report.symbolhistory.symbolhistorycontext import SymbolHistoryContext
from report.symbolhistory.symbolhistoryevent import symbol_history_event_from_option_event
from report.symbolhistory.symbolhistoryevent import symbol_history_event_from_option_order
from report.symbolhistory.symbolhistoryevent import symbol_history_event_from_stock_order
from report.symbolhistory.symbolhistoryevent import SymbolHistoryEvent
from robinhood.lib.account.account import account
from robinhood.lib.account.optionevent.optionevent import OptionEvent
from robinhood.lib.account.optionorder.optionorder import OptionOrder
from robinhood.lib.account.stockorder.stockorder import StockOrder
from robinhood.lib.helpers import format_optional_greek
from robinhood.lib.helpers import format_optional_price
from robinhood.lib.helpers import format_percentage
from robinhood.lib.helpers import format_price
from robinhood.lib.helpers import format_quantity
from robinhood.lib.marketplace.equity import Equity
from robinhood.lib.marketplace.marketplace import marketplace
from robinhood.lib.marketplace.optioninstrument import OptionInstrument
from robinhood.lib.marketplace.optionmarketdata import OptionMarketData
from robinhood.lib.robinhoodclient import rhc


T = Tuple[str, 'OrderedDict[str, Any]', OptionInstrument, Optional[OptionMarketData], int, float]


class SymbolHistory(object):

    def __init__(
        self,
        stock_orders: List[StockOrder],
        option_orders: List[OptionOrder],
        option_events: List[OptionEvent],
    ) -> None:
        self._events_by_symbol = self._load_events_by_symbol(
            stock_orders=stock_orders,
            option_orders=option_orders,
            option_events=option_events,
        )

    @property
    def events_by_symbol(self) -> DefaultDict[str, List[SymbolHistoryEvent]]:
        return self._events_by_symbol

    def get_events(self, symbol: str) -> List[SymbolHistoryEvent]:
        return self._events_by_symbol[symbol]

    def show_current_positions(self) -> None:
        active_symbol = account.active_symbols
        equaities = marketplace.get_equities(symbols=active_symbol)

        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(
                    self.get_current_position_report,
                    symbol=symbol,
                    equity=equaities[symbol],
                )
                for symbol in active_symbol
            ]

        reports_by_symbol: Dict[str, str] = {}
        total_profit = 0.0
        for future in as_completed(futures):
            symbol, report, profit = future.result()
            reports_by_symbol.update({symbol: report})
            total_profit += profit

        color = 'green' if total_profit >= 0 else 'red'
        print(f'Total Profit: {colored(format_price(total_profit), color)}')
        print()

        for symbol in sorted(reports_by_symbol.keys()):
            print(reports_by_symbol[symbol])
            print()

    def get_current_position_report(self, symbol: str, equity: Equity) -> Tuple[str, str, float]:
        events = self.get_events(symbol=symbol)
        if len(events) == 0:
            report = f'No event for {symbol}...'
            return (symbol, report, 0.0)

        event = events[-1]

        # Stock
        symbol_with_price = '{symbol} {last_price} {dod_change}'.format(
            symbol=symbol,
            last_price=format_price(equity.last_price),
            dod_change=format_percentage(equity.dod_percentage),
        )
        symbol_data: OrderedDict[str, Any] = OrderedDict()

        stock_key, stock_details = self._get_current_stock_holding(
            event=event,
            last_price=equity.last_price,
        )
        symbol_data[stock_key] = stock_details

        # Option Buy and Sell
        option_holding_data: List[T] = []
        for buy_or_sell in ['Buy', 'Sell']:
            option_holding_data.extend(
                self._get_current_option_buy_or_sell_holding(
                    event=event,
                    last_price=equity.last_price,
                    buy_or_sell=buy_or_sell,
                ),
            )

        total_profit = 0.0
        exp_dates = sorted(list({
            option_instrument.exp_date
            for _, _, option_instrument, _, _, _ in option_holding_data
        }))
        for exp_date in exp_dates:
            days_to_expire: Optional[int] = None
            current_exp_date_data: List[T] = []
            for option_key, option_details, option_instrument, market_data, quantity, profit in option_holding_data:
                if option_instrument.exp_date == exp_date:
                    current_exp_date_data.append(
                        (
                            option_key,
                            option_details,
                            option_instrument,
                            market_data,
                            quantity,
                            profit,
                        ),
                    )
                    days_to_expire = option_instrument.days_to_expire

            assert days_to_expire is not None
            option_exp_date_key = 'Option Holding w/ Exp Date: {exp_date} ({days_to_expire})'.format(
                exp_date=exp_date,
                days_to_expire=days_to_expire,
            )

            current_option_holding: OrderedDict[str, Any] = OrderedDict()

            # Profit
            subtotal_profit = 0.0
            for _, _, _, _, _, profit in current_exp_date_data:
                subtotal_profit += profit
            color = 'green' if subtotal_profit >= 0 else 'red'
            subtotal_profit_key = f'Total Profit: {colored(format_price(subtotal_profit), color)}'
            current_option_holding[subtotal_profit_key] = {}
            total_profit += subtotal_profit

            # Overall Greeks
            if market_data is not None:
                theta_total = 0.0
                delta_total = 0.0
                gamma_total = 0.0
                vega_total = 0.0
                rho_total = 0.0
                total_quantity = 0
                for _, _, _, market_data, quantity, _ in current_exp_date_data:
                    if market_data is None:
                        continue
                    if market_data.theta is None:
                        continue
                    if market_data.delta is None:
                        continue
                    if market_data.gamma is None:
                        continue
                    if market_data.vega is None:
                        continue
                    if market_data.rho is None:
                        continue

                    theta_total = market_data.theta * quantity
                    delta_total = market_data.delta * quantity
                    gamma_total = market_data.gamma * quantity
                    vega_total = market_data.vega * quantity
                    rho_total = market_data.rho * quantity
                    total_quantity += quantity

                if total_quantity > 0:
                    overall_greek_key = 'Theta: {theta} Delta: {delta} Gamma: {gamma} Vega: {vega} Rho: {rho}'.format(
                        theta=format_optional_greek(theta_total / total_quantity),
                        delta=format_optional_greek(delta_total / total_quantity),
                        gamma=format_optional_greek(gamma_total / total_quantity),
                        vega=format_optional_greek(vega_total / total_quantity),
                        rho=format_optional_greek(rho_total / total_quantity),
                    )
                    current_option_holding[overall_greek_key] = {}

            # Invidual Holding
            for option_key, option_details, _, _, _, _ in sorted(
                current_exp_date_data,
                key=lambda x: x[2].strike_price,
                reverse=True,
            ):
                current_option_holding[option_key] = option_details

            symbol_data[option_exp_date_key] = current_option_holding

        data = {symbol_with_price: symbol_data}

        tree = LeftAligned()
        return (symbol, tree(data), total_profit)

    def _get_current_stock_holding(
        self,
        event: SymbolHistoryEvent,
        last_price: float,
    ) -> 'Tuple[str, OrderedDict[str, Any]]':
        stock_quantity = event.context.stock_holding.quantity
        stock_avg_price = event.context.stock_holding.avg_unit_price
        stock_total_value = stock_quantity * stock_avg_price
        if stock_quantity == 0:
            stock_holding = 'none'
        else:
            stock_holding = '{quantity} x {avg_price} = {total_value}'.format(
                quantity=stock_quantity,
                avg_price=format_price(stock_avg_price),
                total_value=format_price(stock_total_value),
            )
        stock_key = f'Stock Holding: {stock_holding}'
        stock_details: OrderedDict[str, Any] = OrderedDict()

        if stock_quantity > 0:
            current_value = last_price * stock_quantity
            stock_details['Current Value: {value} ({change})'.format(
                value=format_price(current_value),
                change=format_percentage(current_value / stock_total_value - 1),
            )] = {}

        return (stock_key, stock_details)

    def _get_current_option_buy_or_sell_holding(
        self,
        event: SymbolHistoryEvent,
        last_price: float,
        buy_or_sell: str,
    ) -> 'Iterator[T]':
        if buy_or_sell == 'Buy':
            holding_dict = event.context.option_buy_holding
        else:
            holding_dict = event.context.option_sell_holding

        for _, item in holding_dict.items():
            total_base_price = item.avg_unit_price * item.quantity * 100
            if item.option_instrument.instrument_type == 'call':
                if last_price > item.option_instrument.strike_price:
                    diff = last_price - item.option_instrument.strike_price
                    above_or_below = 'above'
                    breached = colored('BREACHED', 'red')
                else:
                    diff = item.option_instrument.strike_price - last_price
                    above_or_below = 'below'
                    breached = colored('NOT BREACHED', 'green')
            else:
                assert item.option_instrument.instrument_type == 'put'
                if last_price < item.option_instrument.strike_price:
                    diff = item.option_instrument.strike_price - last_price
                    above_or_below = 'below'
                    breached = colored('BREACHED', 'red')
                else:
                    diff = last_price - item.option_instrument.strike_price
                    above_or_below = 'above'
                    breached = colored('NOT BREACHED', 'green')
            breached_key = '{breached} - {above_or_below} strike price {diff_dollar} or {diff_percentage}'.format(
                above_or_below=above_or_below,
                diff_dollar=format_price(diff),
                diff_percentage=format_percentage(diff / last_price),
                breached=breached,
            )

            option_sell_key = (
                '{buy_or_sell} {call_or_put} {strike_price}: {quantity} x {avg_price} = {base_price}'.format(
                    buy_or_sell=buy_or_sell,
                    strike_price=format_price(item.option_instrument.strike_price),
                    call_or_put=item.option_instrument.instrument_type.upper(),
                    exp_date=item.option_instrument.exp_date,
                    quantity=item.quantity,
                    avg_price=format_price(item.avg_unit_price),
                    base_price=format_price(total_base_price),
                )
            )
            market_data = item.option_instrument.market_data
            if market_data is None:
                current_price = '-'
                profit_value = 0.0
                profit = '-'
                theta = '-'
                delta = '-'
                gamma = '-'
                vega = '-'
                rho = '-'
            else:
                current_price = format_optional_price(market_data.medium_fill_rate_sell_price)
                if market_data.medium_fill_rate_sell_price is None:
                    profit = '-'
                else:
                    if buy_or_sell == 'Sell':
                        profit_value = (
                            (item.avg_unit_price - market_data.medium_fill_rate_sell_price) * item.quantity * 100
                        )
                        color = 'red' if market_data.medium_fill_rate_sell_price > item.avg_unit_price else 'green'
                    else:
                        assert buy_or_sell == 'Buy'
                        profit_value = (
                            (market_data.medium_fill_rate_sell_price - item.avg_unit_price) * item.quantity * 100
                        )
                        color = 'green' if market_data.medium_fill_rate_sell_price > item.avg_unit_price else 'red'
                    abs_diff = abs(market_data.medium_fill_rate_sell_price - item.avg_unit_price)
                    profit = colored(format_price(abs_diff * item.quantity * 100), color)
                theta = format_optional_greek(market_data.theta)
                delta = format_optional_greek(market_data.delta)
                gamma = format_optional_greek(market_data.gamma)
                vega = format_optional_greek(market_data.vega)
                rho = format_optional_greek(market_data.rho)

            current_price_key = f'Current Price: {current_price}, Profit: {profit}'
            greek_key = f'Theta: {theta} Delta: {delta} Gamma: {gamma} Vega: {vega} Rho: {rho}'

            option_sell_details: OrderedDict[str, Any] = OrderedDict()
            option_sell_details[current_price_key] = {}
            option_sell_details[breached_key] = {}
            option_sell_details[greek_key] = {}

            yield (
                option_sell_key,
                option_sell_details,
                item.option_instrument,
                market_data,
                item.quantity,
                profit_value,
            )

    def show(self, symbol: str, last_n_events: Optional[int] = 10) -> None:
        data = self._gen_ascii_tbl_data(symbol=symbol)
        if last_n_events is None:
            applicable_data = data
        else:
            data_no_head = data[1:]
            applicable_data = [data[0]] + data_no_head[-last_n_events - 1:]
        t = AsciiTable(applicable_data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'center',
            1: 'center',
            2: 'center',
            3: 'right',
            4: 'right',
            5: 'right',
            6: 'right',
            7: 'right',
        }
        print(t.table)

    def show_holdings_by_events(self, symbol: str) -> None:
        events = self.get_events(symbol=symbol)
        for event in events:
            print(event.event_name)
            quantity = event.context.stock_holding.quantity
            avg_price = event.context.stock_holding.avg_unit_price
            print('>>> Stock Holding: {quantity} x {avg_price} = {total_value}'.format(
                quantity=format_quantity(quantity),
                avg_price=format_price(avg_price),
                total_value=format_price(quantity * avg_price),
            ))

            for _, item in event.context.option_buy_holding.items():
                print('>>> Option Buy Holding: {quantity} x {avg_price} = {total_value}'.format(
                    quantity=format_quantity(item.quantity),
                    avg_price=format_price(item.avg_unit_price),
                    total_value=format_price(item.quantity * item.avg_unit_price),
                ))

            for _, item in event.context.option_sell_holding.items():
                print('>>> Option Sell Holding: {quantity} x {avg_price} = {total_value}'.format(
                    quantity=format_quantity(item.quantity),
                    avg_price=format_price(item.avg_unit_price),
                    total_value=format_price(item.quantity * item.avg_unit_price),
                ))

            print()

    def _gen_ascii_tbl_data(self, symbol: str) -> List[List[str]]:
        data = [[
            'Date',
            'Event',
            'Symbol',
            'Quantity',
            'Price',
            '# of Shares',
            'Avg Price',
            'Cumulative Profit',
        ]]

        for event in self._events_by_symbol[symbol]:
            # could be option legs from vertical
            if event.context is None:
                continue

            data.append(
                [
                    event.formatted_ts,
                    event.event_name,
                    event.symbol,
                    format_quantity(event.quantity),
                    format_optional_price(event.unit_price),
                    format_quantity(event.context.stock_holding.quantity),
                    format_price(event.context.stock_holding.avg_unit_price),
                    format_price(event.context.profit),
                ],
            )

        return data

    def _load_events_by_symbol(
        self,
        stock_orders: List[StockOrder],
        option_orders: List[OptionOrder],
        option_events: List[OptionEvent],
    ) -> DefaultDict[str, List[SymbolHistoryEvent]]:
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]] = defaultdict(list)

        self._convert_stock_orders_to_symbol_history_events(
            stock_orders=stock_orders,
            events_by_symbol=events_by_symbol,
        )

        self._convert_option_orders_to_symbol_history_events(
            option_orders=option_orders,
            events_by_symbol=events_by_symbol,
        )

        self._convert_option_events_to_symbol_history_events(
            option_events=option_events,
            events_by_symbol=events_by_symbol,
        )

        for symbol, symbol_events in events_by_symbol.items():
            events_by_symbol[symbol] = sorted(symbol_events, key=lambda e: e.event_ts)

        for symbol, symbol_events in events_by_symbol.items():
            context = SymbolHistoryContext(symbol=symbol)
            for event in symbol_events:
                if event.event_name in ('SHORT PUT', 'SHORT CALL'):
                    assert event.unit_price is not None
                    assert event.option_order.num_of_legs == 1
                    leg = event.option_order.legs[0]
                    context.sell_to_open(
                        symbol=event.symbol,
                        quantity=int(event.quantity) * leg.ratio_quantity,
                        price=abs(event.unit_price),
                        option_instrument=leg.option_instrument,
                    )
                elif event.event_name in ('LONG PUT', 'LONG CALL'):
                    assert event.unit_price is not None
                    assert len(event.option_order.opening_legs) == 1
                    context.buy_to_open(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                        option_instrument=event.option_order.legs[0].option_instrument,
                    )
                elif event.event_name in {'ROLLING CALL', 'ROLLING PUT'}:
                    assert event.unit_price is not None
                    closing_option_instrument = event.option_rolling_order.closing_leg.option_instrument
                    unit_price = context.option_sell_holding[closing_option_instrument.symbol].avg_unit_price
                    context.buy_to_close(
                        symbol=closing_option_instrument.symbol,
                        quantity=int(event.quantity),
                        price=unit_price,
                    )

                    leg = event.option_rolling_order.opening_leg
                    context.sell_to_open(
                        symbol=leg.option_instrument.symbol,
                        quantity=int(event.quantity) * leg.ratio_quantity,
                        price=unit_price - event.unit_price,
                        option_instrument=leg.option_instrument,
                    )
                elif event.event_name == 'SHORT CALL SPREAD':
                    for leg in event.option_order.legs:
                        if leg.side == 'buy':
                            context.buy_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=int(leg.quantity),
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                        else:
                            context.sell_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=int(leg.quantity),
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                elif event.event_name == 'OPTION EXPIRATION':
                    context.option_expire(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                    )
                elif event.event_name == 'BUY TO CLOSE':
                    assert event.unit_price is not None
                    context.buy_to_close(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name == 'SELL TO CLOSE':
                    assert event.unit_price is not None
                    context.sell_to_close(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name == 'CALL ASSIGNMENT':
                    assert event.unit_price is not None
                    context.call_assignment(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name == 'PUT ASSIGNMENT':
                    instrument = event.option_event.option_instrument
                    assert event.unit_price is not None
                    context.put_asignment(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                        strike_price=instrument.strike_price,
                    )
                elif event.event_name == 'OPTION EXERCISE':
                    option_instrument = event.option_event.option_instrument
                    option_avg_price = context.option_buy_holding[event.symbol].avg_unit_price
                    event.unit_price = option_instrument.strike_price + option_avg_price
                    context.call_exercise(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=option_avg_price,
                        strike_price=option_instrument.strike_price,
                    )
                elif event.event_name in ('LIMIT BUY', 'MARKET BUY'):
                    assert event.unit_price is not None
                    context.limit_or_market_buy_stock(
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name in ('LIMIT SELL', 'MARKET SELL'):
                    assert event.unit_price is not None
                    context.limit_or_market_sell_stock(
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name in {
                    'FIG LEAF',
                    'LONG CALL SPREAD',
                    'LONG PUT SPREAD',
                    'SHORT PUT SPREAD',
                    'CUSTOM',
                    'IRON CONDOR',
                    'FIG LEAF CLOSE',
                    'LONG CALL SPREAD CLOSE',
                    'LONG CALL CLOSE',
                    'LONG PUT CLOSE',
                    'LONG PUT SPREAD CLOSE',
                    'SHORT PUT SPREAD CLOSE',
                    'CUSTOM CLOSE',
                    'IRON CONDOR CLOSE',
                    'ROLLING SHORT CALL SPREAD',
                    'ROLLING SHORT PUT SPREAD',
                    'SHORT CALL SPREAD CLOSE',
                }:
                    for leg in event.option_order.legs:
                        quantity = int(event.option_order.processed_quantity) * leg.ratio_quantity
                        if leg.is_buy and leg.is_open:
                            context.buy_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                        elif leg.is_buy and leg.is_close:
                            context.buy_to_close(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                            )
                        elif leg.is_sell and leg.is_open:
                            context.sell_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                        elif leg.is_sell and leg.is_close:
                            context.sell_to_close(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                            )
                else:
                    raise NotImplementedError(event.event_name)

                snapshot = context.take_snapshot()
                event.add_context(symbol_history_context_snapshot=snapshot)

        return events_by_symbol

    def _convert_stock_orders_to_symbol_history_events(
        self,
        stock_orders: List[StockOrder],
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]],
    ) -> None:
        for stock_order in stock_orders:
            if stock_order.instrument_id not in ALL_REVERSED:
                data = rhc.get_url(stock_order.instrument_url)
                for k, v in data.items():
                    print(f'{k}: {v}')

            symbol = ALL_REVERSED[stock_order.instrument_id]
            symbol_history_event = symbol_history_event_from_stock_order(
                symbol=symbol,
                stock_order=stock_order,
            )
            if symbol_history_event is not None:
                events_by_symbol[symbol].append(symbol_history_event)

    def _convert_option_orders_to_symbol_history_events(
        self,
        option_orders: List[OptionOrder],
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]],
    ) -> None:
        for option_order in option_orders:
            symbol = option_order.chain_symbol
            symbol_history_event = symbol_history_event_from_option_order(
                symbol=symbol,
                option_order=option_order,
            )
            if symbol_history_event is not None:
                events_by_symbol[symbol].append(symbol_history_event)

    def _convert_option_events_to_symbol_history_events(
        self,
        option_events: List[OptionEvent],
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]],
    ) -> None:
        for option_event in option_events:
            symbol = option_event.option_instrument.chain_symbol
            symbol_history_event = symbol_history_event_from_option_event(option_event=option_event)
            if symbol_history_event is not None:
                events_by_symbol[symbol].append(symbol_history_event)



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from alphavantage.pricehistory.pricehistoryitem import price_history_item_from_data
from alphavantage.pricehistory.pricehistoryitem import PriceHistoryItem
from alphavantage.pricehistory.pricehistorymetadata import price_history_metadata_from_data
from alphavantage.pricehistory.pricehistorymetadata import PriceHistoryMetadata


class PriceHistory(object):

    def __init__(
        self,
        metadata: PriceHistoryMetadata,
        time_series: 'OrderedDict[str, PriceHistoryItem]',
    ) -> None:
        self._metadata = metadata
        self._time_series = time_series

    @property
    def metadata(self) -> PriceHistoryMetadata:
        return self._metadata

    @property
    def time_series(self) -> 'OrderedDict[str, PriceHistoryItem]':
        return self._time_series


def price_history_from_data(data: Dict[str, Any]) -> PriceHistory:
    metadata = price_history_metadata_from_data(data['Meta Data'])
    time_series_data = data['Time Series (Daily)']
    rows: List[Tuple[str, PriceHistoryItem]] = []
    for ds, data in time_series_data.items():
        rows.append((ds, price_history_item_from_data(data)))

    time_series: OrderedDict[str, PriceHistoryItem] = OrderedDict()
    for ds, price_history_item in sorted(rows, key=lambda x: x[0]):
        time_series[ds] = price_history_item

    return PriceHistory(
        metadata=metadata,
        time_series=time_series,
    )



from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import Dict
from typing import List
from typing import Set

from robinhood.lib.marketplace.equity import Equity
from robinhood.lib.marketplace.equity import equity_from_data
from robinhood.lib.marketplace.optionchain import option_chain_from_data
from robinhood.lib.marketplace.optionchain import OptionChain
from robinhood.lib.robinhoodclient import rhc


class Marketplace(object):

    STOCKS_OPTION_CHAINS_BATCH_SIZE = 64

    def __init__(self) -> None:
        pass

    def get_equities(self, symbols: Set[str]) -> Dict[str, Equity]:
        endpoint = self._get_equities_endpoint(symbols=symbols)
        equities: List[Equity] = rhc.get_objects(
            url=endpoint,
            result_key='results',
            callback=equity_from_data,
        )
        results: Dict[str, Equity] = {}
        for equity in equities:
            results[equity.symbol] = equity

        return results

    def get_option_chains(self, equity_instrument_ids: List[str]) -> Dict[str, OptionChain]:
        num_of_batches = self._get_num_of_option_chains_batches(n=len(equity_instrument_ids))
        with ThreadPoolExecutor() as executor:
            N = self.STOCKS_OPTION_CHAINS_BATCH_SIZE
            futures = [
                executor.submit(
                    self._get_option_chains_one_batch,
                    equity_instrument_ids[i * N:(i + 1) * N],
                )
                for i in range(num_of_batches)
            ]

        option_chains: Dict[str, OptionChain] = {}
        for future in as_completed(futures):
            option_chains.update(future.result())

        return option_chains

    def _get_num_of_option_chains_batches(self, n: int) -> int:
        num_of_batches = n // self.STOCKS_OPTION_CHAINS_BATCH_SIZE
        if n % self.STOCKS_OPTION_CHAINS_BATCH_SIZE > 0:
            num_of_batches += 1

        return num_of_batches

    def _get_option_chains_one_batch(self, equity_instrument_ids: List[str]) -> Dict[str, OptionChain]:
        n = len(equity_instrument_ids)
        assert n <= self.STOCKS_OPTION_CHAINS_BATCH_SIZE

        endpoint = self._get_option_chains_end_point(equity_instrument_ids=equity_instrument_ids)
        option_chains = rhc.get_objects(
            url=endpoint,
            result_key='results',
            callback=option_chain_from_data,
        )
        results: Dict[str, OptionChain] = {}
        for option_chain in option_chains:
            results[option_chain.symbol] = option_chain

        return results

    def _get_equities_endpoint(self, symbols: Set[str]) -> str:
        comma_separated_symbols = ','.join(symbols)
        return f'https://api.robinhood.com/quotes/?symbols={comma_separated_symbols}'

    def _get_option_chains_end_point(self, equity_instrument_ids: List[str]) -> str:
        n = len(equity_instrument_ids)
        assert n <= self.STOCKS_OPTION_CHAINS_BATCH_SIZE

        ids = ','.join(equity_instrument_ids)
        return f'https://api.robinhood.com/options/chains/?equity_instrument_ids={ids}'


marketplace = Marketplace()



from datetime import datetime
from typing import Any

from robinhood.lib.helpers import parse_datetime_with_microsecond


class MarginBalance(object):

    def __init__(
        self,
        start_of_day_dtbp: float,
        outstanding_interest: float,
        gold_equity_requirement: float,
        created_at: datetime,
        updated_at: datetime,
        cash: float,
        cash_held_for_options_collateral: float,
        cash_held_for_orders: float,
        cash_held_for_dividends: float,
        cash_held_for_nummus_restrictions: float,
        cash_held_for_restrictions: float,
        cash_pending_from_options_events: float,
        cash_available_for_withdrawal: float,
        unallocated_margin_cash: float,
        pending_deposit: float,
        uncleared_deposits: float,
        uncleared_nummus_deposits: float,
        day_trade_ratio: float,
        day_trades_protection: bool,
        day_trade_buying_power: float,
        day_trade_buying_power_held_for_orders: float,
        overnight_ratio: float,
        overnight_buying_power: float,
        overnight_buying_power_held_for_orders: float,
        start_of_day_overnight_buying_power: float,
        sma: int,
        settled_amount_borrowed: float,
        unsettled_debit: float,
        unsettled_funds: float,
        crypto_buying_power: float,
    ) -> None:
        self._start_of_day_dtbp = start_of_day_dtbp
        self._outstanding_interest = outstanding_interest
        self._gold_equity_requirement = gold_equity_requirement
        self._created_at = created_at
        self._updated_at = updated_at
        self._cash = cash
        self._cash_held_for_options_collateral = cash_held_for_options_collateral
        self._cash_held_for_orders = cash_held_for_orders
        self._cash_held_for_dividends = cash_held_for_dividends
        self._cash_held_for_nummus_restrictions = cash_held_for_nummus_restrictions
        self._cash_held_for_restrictions = cash_held_for_restrictions
        self._cash_pending_from_options_events = cash_pending_from_options_events
        self._cash_available_for_withdrawal = cash_available_for_withdrawal
        self._unallocated_margin_cash = unallocated_margin_cash
        self._pending_deposit = pending_deposit
        self._uncleared_deposits = uncleared_deposits
        self._uncleared_nummus_deposits = uncleared_nummus_deposits
        self._day_trade_ratio = day_trade_ratio
        self._day_trades_protection = day_trades_protection
        self._day_trade_buying_power = day_trade_buying_power
        self._day_trade_buying_power_held_for_orders = day_trade_buying_power_held_for_orders
        self._overnight_ratio = overnight_ratio
        self._overnight_buying_power = overnight_buying_power
        self._overnight_buying_power_held_for_orders = overnight_buying_power_held_for_orders
        self._start_of_day_overnight_buying_power = start_of_day_overnight_buying_power
        self._sma = sma
        self._settled_amount_borrowed = settled_amount_borrowed
        self._unsettled_debit = unsettled_debit
        self._unsettled_funds = unsettled_funds
        self._crypto_buying_power = crypto_buying_power

    @property
    def overnight_buying_power(self) -> float:
        return self._overnight_buying_power


def margin_balance_from_data(data: Any) -> MarginBalance:
    return MarginBalance(
        start_of_day_dtbp=float(data['start_of_day_dtbp']),
        outstanding_interest=float(data['outstanding_interest']),
        gold_equity_requirement=float(data['gold_equity_requirement']),
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
        cash=float(data['cash']),
        cash_held_for_options_collateral=float(data['cash_held_for_options_collateral']),
        cash_held_for_orders=float(data['cash_held_for_orders']),
        cash_held_for_dividends=float(data['cash_held_for_dividends']),
        cash_held_for_nummus_restrictions=float(data['cash_held_for_nummus_restrictions']),
        cash_held_for_restrictions=float(data['cash_held_for_restrictions']),
        cash_pending_from_options_events=float(data['cash_pending_from_options_events']),
        cash_available_for_withdrawal=float(data['cash_available_for_withdrawal']),
        unallocated_margin_cash=float(data['unallocated_margin_cash']),
        pending_deposit=float(data['pending_deposit']),
        uncleared_deposits=float(data['uncleared_deposits']),
        uncleared_nummus_deposits=float(data['uncleared_nummus_deposits']),
        day_trade_ratio=float(data['day_trade_ratio']),
        day_trades_protection=data['day_trades_protection'],
        day_trade_buying_power=float(data['day_trade_buying_power']),
        day_trade_buying_power_held_for_orders=float(data['day_trade_buying_power_held_for_orders']),
        overnight_ratio=float(data['overnight_ratio']),
        overnight_buying_power=float(data['overnight_buying_power']),
        overnight_buying_power_held_for_orders=float(data['overnight_buying_power_held_for_orders']),
        start_of_day_overnight_buying_power=float(data['start_of_day_overnight_buying_power']),
        sma=int(data['sma']),
        settled_amount_borrowed=float(data['settled_amount_borrowed']),
        unsettled_debit=float(data['unsettled_debit']),
        unsettled_funds=float(data['unsettled_funds']),
        crypto_buying_power=float(data['crypto_buying_power']),
    )



from typing import Any
from typing import Dict


class PriceHistoryMetadata(object):

    def __init__(
        self,
        information: str,
        symbol: str,
        output_size: str,
        last_refreshed: str,
        timezone: str,
    ) -> None:
        self._symbol = symbol
        self._information = information
        self._output_size = output_size
        self._last_refreshed = last_refreshed
        self._timezone = timezone

    @property
    def information(self) -> str:
        return self._information


def price_history_metadata_from_data(data: Dict[str, Any]) -> PriceHistoryMetadata:
    return PriceHistoryMetadata(
        information=data['1. Information'],
        symbol=data['2. Symbol'],
        last_refreshed=data['3. Last Refreshed'],
        output_size=data['4. Output Size'],
        timezone=data['5. Time Zone'],
    )



from collections import defaultdict
from datetime import datetime
from typing import DefaultDict
from typing import List
from typing import Set

from terminaltables import AsciiTable

from report.symbolhistory.symbolhistory import SymbolHistory
from robinhood.lib.helpers import format_price
from robinhood.lib.helpers import get_month
from robinhood.lib.helpers import get_week


class AccountPerformance(object):

    def __init__(self, symbol_history: SymbolHistory) -> None:
        self._symbol_history = symbol_history

    def show(self, interval: str) -> None:
        assert interval in {'Week', 'Month', 'All Time'}
        print(self._gen_reports(interval=interval))

    def _gen_reports(self, interval: str) -> str:
        data = self._gen_data(interval=interval)
        t = AsciiTable(data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'center',
            1: 'right',
            2: 'right',
            3: 'right',
        }

        return str(t.table)

    def _gen_data(self, interval: str) -> List[List[str]]:
        if interval == 'Week':
            event_date_func = get_week
        elif interval == 'Month':
            event_date_func = get_month
        else:
            assert interval == 'All Time'

            def event_date_func(d: datetime) -> str:
                return '-'

        data = [[
            interval,
            'Settled Profit from Stock',
            'Settled Profit from Option',
            'Total Settled Profit',
        ]]

        stock_profit: DefaultDict[str, float] = defaultdict(float)
        option_profit: DefaultDict[str, float] = defaultdict(float)

        for _, events in self._symbol_history.events_by_symbol.items():
            for i in range(1, len(events)):
                previous_event = events[i - 1]
                current_event = events[i]
                diff = current_event.context.profit - previous_event.context.profit
                if diff != 0:
                    event_date = event_date_func(current_event.event_ts)

                    if current_event.event_name in {
                        'BUY TO CLOSE',
                        'CUSTOM',
                        'CUSTOM CLOSE',
                        'FIG LEAF CLOSE',
                        'IRON CONDOR CLOSE',
                        'LONG CALL CLOSE',
                        'LONG CALL SPREAD CLOSE',
                        'LONG PUT CLOSE',
                        'LONG PUT SPREAD CLOSE',
                        'OPTION EXPIRATION',
                        'ROLLING SHORT CALL SPREAD',
                        'SHORT CALL SPREAD CLOSE',
                        'SHORT PUT SPREAD CLOSE',
                    }:
                        option_profit[event_date] += diff
                    elif current_event.event_name in {
                        'CALL ASSIGNMENT',
                        'LIMIT SELL',
                        'MARKET SELL',
                    }:
                        stock_profit[event_date] += diff
                    else:
                        raise Exception(f'Unrecognized event name: {current_event.event_name}')

        all_dates: Set[str] = set()
        for event_date in stock_profit.keys():
            all_dates.add(event_date)
        for event_date in option_profit.keys():
            all_dates.add(event_date)

        for event_date in sorted(list(all_dates)):
            total_profit = stock_profit[event_date] + option_profit[event_date]
            data.append([
                event_date,
                format_price(stock_profit[event_date]),
                format_price(option_profit[event_date]),
                format_price(total_profit),
            ])

        return data



from typing import Optional

from robinhood.lib.marketplace.optioninstrument import OptionInstrument


class SymbolHistoryContextHolding(object):

    def __init__(
        self,
        symbol: str,
        quantity: int,
        avg_unit_price: float,
        option_instrument: Optional[OptionInstrument],
    ) -> None:
        self._symbol = symbol
        self._quantity = quantity
        self._avg_unit_price = avg_unit_price
        self._option_instrument = option_instrument

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def quantity(self) -> int:
        return self._quantity

    @property
    def avg_unit_price(self) -> float:
        if self._quantity == 0:
            return 0.0

        return self._avg_unit_price

    @property
    def stock_cost_basis(self) -> float:
        assert self._option_instrument is None
        return self.quantity * self.avg_unit_price

    @property
    def option_instrument(self) -> OptionInstrument:
        assert self._option_instrument is not None
        return self._option_instrument

    @property
    def option_buy_profit(self) -> float:
        assert self._option_instrument is not None
        market_data = self._option_instrument.market_data
        assert market_data is not None
        return (market_data.adjusted_mark_price - self.avg_unit_price) * self.quantity * 100

    @property
    def option_buy_value(self) -> float:
        assert self._option_instrument is not None
        market_data = self._option_instrument.market_data
        assert market_data is not None
        return market_data.adjusted_mark_price * self.quantity * 100



from setuptools import find_packages
from setuptools import setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='faith',
    version='0.0.28',
    author='Thomas Wang',
    author_email='w@datability.io',
    description='the faith API for investment portfolio management',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/datability-io/faith',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'asciitree==0.3.3',
        'pytz==2019.2',
        'redis==3.3.8',
        'requests==2.22.0',
        'termcolor==1.1.0',
        'terminaltables==3.1.0',
    ],
    package_data={'faith': ['py.typed']},
)



import os
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from robinhood.lib.account.stockorder.stockorderexecution import stock_order_execution_from_data
from robinhood.lib.account.stockorder.stockorderexecution import StockOrderExecution
from robinhood.lib.helpers import parse_datetime_with_microsecond
from robinhood.lib.helpers import parse_optional_float


class StockOrder(object):

    def __init__(
        self,
        order_id: str,
        order_type: str,
        order_url: str,
        side: str,
        state: str,
        trigger: str,
        instrument_url: str,
        price: Optional[float],
        average_price: Optional[float],
        quantity: float,
        cumulative_quantity: float,
        executions: List[StockOrderExecution],
        stop_price: Optional[float],
        last_trail_price: Optional[float],
        fees: float,
        ref_id: str,
        extended_hours: bool,
        time_in_force: str,
        response_category: str,
        account_url: str,
        override_dtbp_checks: bool,
        override_day_trade_checks: bool,
        position_url: str,
        last_transaction_at: datetime,
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._order_id = order_id
        self._order_type = order_type
        self._order_url = order_url
        self._side = side
        self._state = state
        self._trigger = trigger
        self._instrument_url = instrument_url
        self._price = price
        self._average_price = average_price
        self._quantity = quantity
        self._cumulative_quantity = cumulative_quantity
        self._executions = executions
        self._stop_price = stop_price
        self._last_trail_price = last_trail_price
        self._fees = fees
        self._ref_id = ref_id
        self._extended_hours = extended_hours
        self._time_in_force = time_in_force
        self._response_category = response_category
        self._account_url = account_url
        self._override_dtbp_checks = override_dtbp_checks
        self._override_day_trade_checks = override_day_trade_checks
        self._position_url = position_url
        self._last_transaction_at = last_transaction_at
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def order_type(self) -> str:
        return self._order_type

    @property
    def side(self) -> str:
        return self._side

    @property
    def state(self) -> str:
        return self._state

    @property
    def instrument_id(self) -> str:
        return os.path.basename(os.path.normpath(self._instrument_url))

    @property
    def instrument_url(self) -> str:
        return self._instrument_url

    @property
    def average_price(self) -> Optional[float]:
        return self._average_price

    @property
    def quantity(self) -> float:
        return self._quantity

    @property
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    def symbol_event_name(self) -> str:
        return f'{self._order_type} {self._side}'.upper()


def stock_order_from_data(data: Dict[str, Any]) -> StockOrder:
    executions = [
        stock_order_execution_from_data(data=row)
        for row in data['executions']
    ]
    return StockOrder(
        order_id=data['id'],
        order_type=data['type'],
        order_url=data['url'],
        side=data['side'],
        state=data['state'],
        trigger=data['trigger'],
        instrument_url=data['instrument'],
        price=parse_optional_float(data['price']),
        average_price=parse_optional_float(data['average_price']),
        quantity=float(data['quantity']),
        cumulative_quantity=float(data['cumulative_quantity']),
        executions=executions,
        stop_price=parse_optional_float(data['stop_price']),
        last_trail_price=parse_optional_float(data['last_trail_price']),
        fees=float(data['fees']),
        ref_id=data['ref_id'],
        extended_hours=data['extended_hours'],
        time_in_force=data['time_in_force'],
        response_category=data['response_category'],
        account_url=data['account'],
        override_dtbp_checks=data['override_dtbp_checks'],
        override_day_trade_checks=data['override_day_trade_checks'],
        position_url=data['position'],
        last_transaction_at=parse_datetime_with_microsecond(data['last_transaction_at']),
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )



from datetime import datetime
from typing import Optional

from report.symbolhistory.symbolhistorycontextsnapshot import SymbolHistoryContextSnapshot
from robinhood.lib.account.optionevent.optionevent import OptionEvent
from robinhood.lib.account.optionorder.optionorder import OptionOrder
from robinhood.lib.account.optionorder.optionrollingorder import OptionRollingOrder
from robinhood.lib.account.stockorder.stockorder import StockOrder


class SymbolHistoryEvent(object):

    def __init__(
        self,
        event_ts: datetime,
        event_name: str,
        symbol: str,
        quantity: float,
        unit_price: Optional[float],
        stock_order: Optional[StockOrder] = None,
        option_order: Optional[OptionOrder] = None,
        option_rolling_order: Optional[OptionRollingOrder] = None,
        option_event: Optional[OptionEvent] = None,
    ):
        self._event_ts = event_ts
        self._event_name = event_name
        self._symbol = symbol
        self._quantity = quantity
        self._unit_price = unit_price
        self._stock_order = stock_order
        self._option_order = option_order
        self._option_rolling_order = option_rolling_order
        self._option_event = option_event

        self._context: Optional[SymbolHistoryContextSnapshot] = None

    @property
    def event_name(self) -> str:
        return self._event_name

    @property
    def event_ts(self) -> datetime:
        return self._event_ts

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def quantity(self) -> float:
        return self._quantity

    @property
    def unit_price(self) -> Optional[float]:
        return self._unit_price

    @unit_price.setter
    def unit_price(self, unit_price: float) -> None:
        """
        Why do we need to set unit price?
        """
        self._unit_price = unit_price

    @property
    def stock_order(self) -> StockOrder:
        assert self._stock_order is not None
        return self._stock_order

    @property
    def option_order(self) -> OptionOrder:
        assert self._option_order is not None
        return self._option_order

    @property
    def option_event(self) -> OptionEvent:
        assert self._option_event is not None
        return self._option_event

    @property
    def option_rolling_order(self) -> OptionRollingOrder:
        assert self._option_rolling_order is not None
        return self._option_rolling_order

    @property
    def context(self) -> SymbolHistoryContextSnapshot:
        assert self._context is not None
        return self._context

    @property
    def formatted_ts(self) -> str:
        return self.event_ts.strftime('%m/%d/%Y %H:%M')

    def add_context(
        self,
        symbol_history_context_snapshot: SymbolHistoryContextSnapshot,
    ) -> None:
        self._context = symbol_history_context_snapshot


def symbol_history_event_from_stock_order(symbol: str, stock_order: StockOrder) -> Optional[SymbolHistoryEvent]:
    if stock_order.state != 'filled':
        return None

    assert stock_order.average_price is not None

    return SymbolHistoryEvent(
        event_ts=stock_order.updated_at,
        event_name=stock_order.symbol_event_name,
        symbol=symbol,
        quantity=stock_order.quantity,
        unit_price=stock_order.average_price,
        stock_order=stock_order,
    )


def symbol_history_event_from_option_order(
    symbol: str,
    option_order: OptionOrder,
) -> Optional[SymbolHistoryEvent]:
    if not (
        option_order.state == 'filled'
        or (option_order.state == 'cancelled' and option_order.processed_quantity > 0)
    ):
        return None

    option_rolling_order = option_order.rolling_order
    if option_rolling_order is not None:
        call_or_put = option_rolling_order.call_or_put.upper()
        event_name = f'ROLLING {call_or_put}'
        symbol = option_rolling_order.symbol
        quantity = option_order.processed_quantity
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=event_name,
            symbol=symbol,
            quantity=quantity,
            unit_price=option_order.price,
            option_rolling_order=option_rolling_order,
        )

    if option_order.is_rolling_short_spread:
        call_or_put = option_order.legs[0].option_instrument.instrument_type.upper()
        event_name = f'ROLLING SHORT {call_or_put} SPREAD'
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=event_name,
            symbol=option_order.concatenated_symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.is_buy_to_close:
        event_name = 'BUY TO CLOSE'
        symbol = option_order.legs[0].option_instrument.symbol
        quantity = option_order.processed_quantity
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=event_name,
            symbol=symbol,
            quantity=quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.opening_strategy in {
        'short_call',
        'short_put',
        'long_call',
        'long_put',
    }:
        assert option_order.opening_strategy_name is not None
        assert len(option_order.opening_legs) == 1
        assert option_order.num_of_legs == 1
        event_name = option_order.opening_strategy_name
        symbol = option_order.opening_legs[0].option_instrument.symbol
        quantity = option_order.processed_quantity
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=event_name,
            symbol=symbol,
            quantity=quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.opening_strategy in {
        'long_call_spread',
        'long_put_spread',
        'short_call_spread',
        'short_put_spread',
    }:
        assert option_order.opening_strategy_name is not None
        assert option_order.num_of_legs == 2
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=option_order.opening_strategy_name,
            symbol=option_order.spread_symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.closing_strategy in {
        'short_call_spread',
    }:
        assert option_order.closing_strategy is not None
        assert option_order.num_of_legs == 2
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=f'{option_order.closing_strategy_name} CLOSE',
            symbol=option_order.spread_symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.is_fig_leaf_open:
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name='FIG LEAF',
            symbol=option_order.fig_leaf_symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.is_fig_leaf_close:
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name='FIG LEAF CLOSE',
            symbol=option_order.fig_leaf_symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.closing_strategy in {
        'long_call_spread',
        'long_put_spread',
        'short_put_spread',
    }:
        if option_order.price is None:
            unit_price = 0.0
            for leg in option_order.legs:
                multiplier = -1 if leg.side == 'buy' else 1
                for ex in leg.executions:
                    unit_price += multiplier * ex.price
        else:
            unit_price = option_order.price

        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=f'{option_order.closing_strategy_name} CLOSE',
            symbol=option_order.spread_symbol,
            quantity=option_order.processed_quantity,
            unit_price=unit_price,
            option_order=option_order,
        )

    if option_order.closing_strategy in {
        'long_call',
        'long_put',
    }:
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=f'{option_order.closing_strategy_name} CLOSE',
            symbol=option_order.legs[0].option_instrument.symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.opening_strategy in {
        'custom',
        'iron_condor',
    }:
        assert option_order.opening_strategy_name is not None

        event_name = option_order.opening_strategy_name
        symbol = option_order.concatenated_symbol
        quantity = option_order.processed_quantity
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=event_name,
            symbol=symbol,
            quantity=quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    if option_order.closing_strategy in {
        'custom',
        'iron_condor',
    }:
        return SymbolHistoryEvent(
            event_ts=option_order.updated_at,
            event_name=f'{option_order.closing_strategy_name} CLOSE',
            symbol=option_order.concatenated_symbol,
            quantity=option_order.processed_quantity,
            unit_price=option_order.price,
            option_order=option_order,
        )

    print(
        f'{option_order.order_id}: {option_order.chain_symbol} '
        f'Open: {option_order.opening_strategy}: '
        f'{option_order.num_of_legs} {option_order.legs[0].side} {option_order.legs[0].position_effect}',
    )
    print(
        f'{option_order.order_id}: {option_order.chain_symbol} '
        f'Close: {option_order.closing_strategy}: '
        f'{option_order.direction}',
    )
    raise Exception('Unrecognized Option Order')


def symbol_history_event_from_option_event(option_event: OptionEvent) -> Optional[SymbolHistoryEvent]:
    if option_event.state not in {'confirmed', 'pending'}:
        return None

    if option_event.event_type == 'expiration':
        return SymbolHistoryEvent(
            event_ts=option_event.created_at,
            event_name=f'option {option_event.event_type}'.upper(),
            symbol=option_event.option_instrument.symbol,
            quantity=option_event.quantity,
            unit_price=0.0,
            option_event=option_event,
        )

    if option_event.event_type == 'assignment':
        if option_event.direction == 'credit':
            option_type = 'call'
        else:
            option_type = 'put'

        return SymbolHistoryEvent(
            event_ts=option_event.created_at,
            event_name=f'{option_type} {option_event.event_type}'.upper(),
            symbol=option_event.option_instrument.symbol,
            quantity=option_event.quantity,
            unit_price=option_event.option_instrument.strike_price,
            option_event=option_event,
        )

    return SymbolHistoryEvent(
        event_ts=option_event.created_at,
        event_name=f'option {option_event.event_type}'.upper(),
        symbol=option_event.option_instrument.symbol,
        quantity=option_event.quantity,
        unit_price=option_event.option_instrument.strike_price,
        option_event=option_event,
    )



from typing import Dict


class MinTicks(object):
    def __init__(
        self,
        cutoff_price: float,
        below_tick: float,
        above_tick: float,
    ) -> None:
        self._cutoff_price = cutoff_price
        self._below_tick = below_tick
        self._above_tick = above_tick

    @property
    def cutoff_price(self) -> float:
        return self._cutoff_price

    @property
    def below_tick(self) -> float:
        return self._below_tick

    @property
    def above_tick(self) -> float:
        return self._above_tick

    def get_min_tick_price_round_up(self, price: float) -> float:
        return self._get_min_tick_price(price=price, up_or_down='up')

    def get_min_tick_price_round_down(self, price: float) -> float:
        return self._get_min_tick_price(price=price, up_or_down='down')

    def _get_min_tick_price(self, price: float, up_or_down: str) -> float:
        if up_or_down not in {'up', 'down'}:
            raise ValueError(f'Incorret parameter `up_or_down`: {up_or_down}')

        if price == self._cutoff_price:
            return price

        if price < self._cutoff_price:
            tick = self._below_tick
        else:
            tick = self._above_tick

        price100 = price * 100
        tick100 = tick * 100

        if price100 % tick100 == 0:
            return price

        _n = price100 // tick100
        if up_or_down == 'up':
            _n += 1

        return _n * tick100 / 100


def min_ticks_from_data(data: Dict[str, str]) -> MinTicks:
    return MinTicks(
        cutoff_price=float(data['cutoff_price']),
        below_tick=float(data['below_tick']),
        above_tick=float(data['above_tick']),
    )
